use glutin::{
    dpi::{PhysicalPosition, PhysicalSize},
    event_loop::EventLoop,
    window::Window as GlutinWindow,
    window::WindowBuilder,
    Api, ContextBuilder, ContextWrapper, GlProfile, GlRequest, PossiblyCurrent,
};
use log::info;

use crate::Error;

pub struct Window {
    context: ContextWrapper<PossiblyCurrent, GlutinWindow>,
}

impl Window {
    pub fn new<T>(event_loop: &EventLoop<T>) -> Result<Self, Error> {
        let mut multisampling = [16u16, 8, 4, 2, 1, 0].iter();
        let context = loop {
            let multisampling = match multisampling.next() {
                Some(multisampling) => multisampling,
                None => return Err(Error::CreateContext),
            };

            info!("Create OpenGL context (multisampling={})", multisampling);

            let window_builder = WindowBuilder::new()
                .with_title("space-crush")
                .with_visible(true)
                .with_inner_size(PhysicalSize::new(WINDOW_WIDTH, WINDOW_HEIGHT));
            let ret = ContextBuilder::new()
                .with_double_buffer(Some(true))
                .with_hardware_acceleration(Some(true))
                .with_pixel_format(24, 8)
                .with_multisampling(*multisampling)
                .with_vsync(false)
                .with_gl(GlRequest::Specific(Api::OpenGl, (4, 5)))
                .with_gl_profile(GlProfile::Core)
                .build_windowed(window_builder, event_loop);

            if let Ok(context) = ret {
                break context;
            }
        };

        let window = context.window();
        let window_size = window.outer_size();
        let monitor_size = window.current_monitor().unwrap().size();
        window.set_outer_position(PhysicalPosition::new(
            (monitor_size.width - window_size.width) / 2,
            (monitor_size.height - window_size.height) / 2,
        ));

        let context = unsafe { context.make_current().unwrap() };
        gl::load_with(|s| context.get_proc_address(s));

        gl::disable(gl::CULL_FACE);
        gl::disable(gl::DEPTH_TEST);
        gl::pixel_store_i(gl::UNPACK_ALIGNMENT, 1);

        Ok(Self { context })
    }

    pub fn swap_buffers(&self) -> Result<(), Error> {
        self.context.swap_buffers()?;

        Ok(())
    }
}

const WINDOW_WIDTH: u32 = 1280;
const WINDOW_HEIGHT: u32 = 720;
