use std::any::type_name;
use std::iter::Iterator;

use glc::{
    shader::{Program, Shader, Type},
    texture::{Data, Target, Texture},
};
use shred::{Fetch, FetchMut, Resource};
use shrev::{Event, EventChannel, ReaderId};
use specs::World;

use crate::{misc::Vfs, Error};

pub trait WorldHelper {
    fn resource<R>(&self) -> Result<Fetch<R>, Error>
    where
        R: Resource;

    fn resource_mut<R>(&self) -> Result<FetchMut<R>, Error>
    where
        R: Resource;

    fn register_event_reader<E>(&self) -> Result<ReaderId<E>, Error>
    where
        E: Event;

    fn load_program<I>(&self, iter: I) -> Result<Program, Error>
    where
        I: IntoIterator<Item = (Type, &'static str)>;

    fn load_texture(&self, path: &str) -> Result<Texture, Error>;
}

impl WorldHelper for World {
    fn resource<R>(&self) -> Result<Fetch<R>, Error>
    where
        R: Resource,
    {
        self.try_fetch::<R>()
            .ok_or_else(|| Error::ResourceNotRegistered(type_name::<R>()))
    }

    fn resource_mut<R>(&self) -> Result<FetchMut<R>, Error>
    where
        R: Resource,
    {
        self.try_fetch_mut::<R>()
            .ok_or_else(|| Error::ResourceNotRegistered(type_name::<R>()))
    }

    fn register_event_reader<E>(&self) -> Result<ReaderId<E>, Error>
    where
        E: Event,
    {
        Ok(self.resource_mut::<EventChannel<E>>()?.register_reader())
    }

    fn load_program<I>(&self, iter: I) -> Result<Program, Error>
    where
        I: IntoIterator<Item = (Type, &'static str)>,
    {
        let vfs = self.fetch::<Vfs>();

        Program::from_shaders_result(iter.into_iter().map(|(t, p)| {
            let mut file = vfs.join(p)?.open_file()?;
            let shader = Shader::from_reader(t, &mut file)?;

            Ok(shader)
        }))
    }

    fn load_texture(&self, path: &str) -> Result<Texture, Error> {
        let vfs = self.fetch::<Vfs>();

        let mut file = vfs.join(path)?.open_file()?;
        let mut data = Data::from_reader(&mut file)?;
        data.convert_to(data.format().info().format_gl_support)?;

        let mut texture = Texture::new(Target::Texture2D)?;
        texture.upload(&data, true)?;

        Ok(texture)
    }
}
